#####################################################################
#:CQuest.pm
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2011. All Rights Reserved.
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

=head1 NAME

CQCC::CQuest - Base class to manage communications with ClearQuest

=head1 CLASS DESCRIPTION

This class provides an abstract base class for all ClearQuest
interfaces. It handles user logon information and storage, defining
class parameters such as query definitions, and provides standard unit
tests. A subclass such as CQCC::CQWeb must implement methods to logon,
post commands, format queries and modify entity data.

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

######################################################################
#::Package CQCC::CQuest
######################################################################

package CQCC::CQuest;

use strict;
use CQCC::IOBase;
use CQCC::Parm;
use CQCC::CQSchema;
use CQCC::AssocList;
use CQCC::Timer; 

sub Version { return '2.2' };

######################################################################
#::Class Globals 
######################################################################

my %performQueryCache = ();   # Used to store query results

# Should be consolidated with schemas already defined in CQPerlAPI
my $csetSchema2;
my $perl5006 = ($] ge "5.006" ? 1 : 0);
my $CQCCext701 = 0;

######################################################################
#::function ClassInit

=over 6

=item C<ClassInit()>

This class function is automatically invoked to define class 
parameters using CQCC::Parm objects. Current parameters include: 

=over 4

=item C<USER>

An internal parameter initialized during logon which can be 
referenced in other variables such as Query filters, etc. 

=item C<CQUSER>

An internal parameter like USER. If CQUSER is defined, it can be 
used to override the value of USER (the CQ login name) to provide
the internal CQ user name needed by Query filters.
 
=item C<REPLICA_NAME>

An internal parameter initialized during queries which can be 
use to reference the replica name of a MultiSite database. 

=item C<ClearQuestInterface>

Name of CQuest subclass to use for this session, used by
IOClassFactory(). Set to 'CQCC::CQWebJava' by default.

=item C<CQCC_DATABASE>

Name of default database.

=item C<CQCC_ENTITY>

Name of default entity.

=item C<CQCC_MULTISITE>

Switch to enable checking for ClearQuest Multisite mastership
restrictions before making associations. Set to 'FALSE' by 
default; change to 'TRUE' to enable. 

=item C<CQCC_REPLICA_NAME>

Default value for ClearQuest database replica name.
Only used when CQCC_MULTISITE is enabled and there is no 
way to automatically obtain the same information through the API.

=back

=back

=cut

######################################################################

sub ClassInit { 
   # Internal only, not Config
    &DefineParm(NAME=>"USER", PCLASS=>"CQCC::CQuest", SRC_INT=>1);

    &DefineParm(NAME=>"CQUSER", PCLASS=>"CQCC::CQuest", SRC_INT=>1);

    &DefineParm(NAME=>"REPLICA_NAME", PCLASS=>"CQCC::CQuest", 
                SRC_INT=>1);

    &SetParmDefaults(PCLASS=>"CQCC::CQuest", SRC_CFG=>1);

    &DefineParm(NAME=>"ClearQuestInterface", DEFAULT=>"CQCC::CQWebJava",
                SRC_CMD=>1);

    &DefineParm(NAME=>"CQCC_DATABASE", SRC_EV=>1, SRC_CACHE=>1,
                SRC_USER_PREF=>1);

    &DefineParm(NAME=>"CQCC_ENTITY", SRC_EV=>1, SRC_CACHE=>1,
                SRC_USER_PREF=>1);

    &DefineParm(NAME=>"CQCC_QUERY_NAME", SRC_EV=>1, SRC_CACHE=>1,
                SRC_USER_PREF=>1, DEFAULT=>"STANDARD");

    &DefineParm(NAME=>"CQCC_QUERY_TYPE", SRC_EV=>1, SRC_CACHE=>1,
                SRC_USER_PREF=>1, DEFAULT=>"LOCAL", 
                RANGE_LIST=> [ "LOCAL", "CQ" ]);

    &DefineParm(NAME=>"CQCC_MULTISITE", DEFAULT=>"FALSE",
                RANGE_LIST => ["FALSE", "TRUE"]);

    &DefineParm(NAME=>"CQCC_REPLICA_NAME", SRC_EV=>1);

    # Internal variables.

    &DefineParm(NAME=>"CQCC_QUERY_OID", DEFAULT=>"FALSE",
                RANGE_LIST => ["FALSE", "TRUE"]);

    &SetParmDefaults();
}

&ClassInit();

######################################################################
#::function ClassTest

=over 6

=item C<ClassTest(@args)>

This class function uses sets up a local environment for testing;
calls IOClassFactory() to get an instance of the desired subclass; 
and invokes the object's SelfTest method for testing. 

=back

=cut

######################################################################

sub ClassTest { 
    my @args = @_;
    my $cq = &IOClassFactory("ClearQuestInterface");
    return $cq->SelfTest(@args);
}

######################################################################
#::function SetCQUser

=over 6

=item C<SetCQUser($cqusername)>

This class method sets PARM CQUSER to cqusername and 
writes cqusername to the parmfile for future sessions to use. 
It returns 1 if no error, 0 otherwise.   

=back

=cut

######################################################################

sub SetCQUser {
    my ($self, $cqusername) = @_;
    my $result =0;
    
    &ForceParm("CQUSER", $cqusername); 
    &IODebugPrint("DETAILED", "In CQuest::SetCQUser CQUSER = $cqusername \n");
    my $parmfilename = $self->{PARMFILE};
    if ($ENV{"CQCC_TEST_CQUSER"} && 
        $ENV{"CQCC_TEST_CQUSER"} ne $cqusername) {
        IOMessage("ERROR", "ATE: CQCC_TEST_CQUSER(" . 
             $ENV{"CQCC_TEST_CQUSER"} . 
             ") is different from returned value ($cqusername)\n");
    }
    if (-s $parmfilename) {
	my ($answer) = &IOFileRead($parmfilename,0);
	my ($user, $pass, $cquser) = split(" ", $answer);
	if ( defined $user) {
	    if ( ($cquser eq undef )|| ($cquser ne $cqusername )) {
		&IOFileWrite($parmfilename, "$user $pass $cqusername" );
	    }           
	    $result =1;
	}
    }
    &IODebugPrint("DETAILED", "CQuest::SaveCQUser returns $result\n");    
    return $result; 
}

######################################################################
#::function GetCQUser

=over 6

=item C<GetCQUser()>

This class method reads PARM CQUSER and returns it to the caller.  
The returned value can be "undef". 

=back

=cut

######################################################################

sub GetCQUser {   
    return &GetParm("CQUSER"); 
}

######################################################################
#::method Connect

=over 6

=item C<Connect($return)>

This method repeatedly tries to logon to the ClearQuest database
until the user succeeds or gives up trying. If the $return argument
is non-zero, then instead of exiting on failure the method will 
return a zero.

=back

=cut

######################################################################

sub Connect {
    my ($self, $return) = @_;
    my $result = 0;

    while (! $result) {
       my $t1 = &TimerStart("CQuest::Logon");
       my $db = &GetParm("CQCC_DATABASE");
       if ($self->Logon()) {
           $self->{DATABASE} = $db; 
           $result = 1;
       } else {
           # CQ logon failed. If it is in ATE test, don't prompt 
           # for logon retry.
           if ($ENV{"CQCC_TEST_USER"} || $ENV{"CQCC_CQ_NOLOGIN"}) {
              &IOMessage("FATAL", ["CQ_E_TEST_LOGIN_FAILED"]);
           }
           my ($abort,$answer) =
              &IOPromptYesNo(["CQ_E_LOGIN_FAILED", $db ]);

           # Force uncheckout if necessary
           if ($abort || $answer ne 1) {
               if ($return) {
                   return 0;
               }
               &IOExitHandler(1);
           } else {
               $self->ClearLogonParms();
          }
       }
       &TimerStop($t1);
    }
    # If successful login and no replica name set, use variable. 
    if ((&GetParm("CQCC_MULTISITE") eq "TRUE") && 
           ! defined &GetParm("REPLICA_NAME")) {
       &ForceParm("REPLICA_NAME", &GetParm("CQCC_REPLICA_NAME"));
       if (! defined &GetParm("REPLICA_NAME")) {
           &IOMessage("FATAL", ["CQ_E_UNKNOWN_REPLICA_NAME"]);
       }
    }
    return $result;
}

######################################################################
#::method EditAssocs

=over 6

=item C<EditAssocs($VOList, $addAssocs, $delAssocs)>

This method takes information about the ClearCase version object being
associated to and two AssocLists, one listing the new associations
to be made, and the other, the list of associations to be deleted.

The VOList is expected to be a string with a comma separated list of
pathname, vobid, objectID, and old objectID.

If the system fails to successfully make the new additions, they are
all deleted to prepare for the trigger failing gracefully.

=back

=cut

######################################################################

sub EditAssocs {
    my ($self, $voList, $addAssocs, $delAssocs) = @_;
    my ($dbname, $entity);
    my @entityDBlist;
    my $result = 1;

    my $t1 = &TimerStart("CQuest::EditAssocs");
    if ($addAssocs) {
        @entityDBlist = $addAssocs->GetEntityDBList();
    }
    if ($delAssocs) {
        @entityDBlist = $delAssocs->GetEntityDBList(@entityDBlist);
    }
    if ($#entityDBlist > 0) { 
        &IOMessage("ERROR", ["CQ_E_MIXED_ENTITIES",
            join(", ",@entityDBlist)]);
        return;
    } 
    for (@entityDBlist) { 
        my ($entity, $dbname) = split(":",$_);
        &IODebugPrint("BASIC", "CQuest::EditAssocs db ($dbname) " . 
            "entity ($entity) volist ($voList)");
        my $r = $self->_EditAssocs($dbname, $entity, $voList, 
            $addAssocs, $delAssocs);
        if ($r == 0) { 
            $result = 0;
        }
    }
    &TimerStop($t1);
    return $result;
}
    
######################################################################
#::method ClearLogonParms

=over 6

=item C<ClearLogonParms()>

Method clears cached user name and password to logon to ClearQuest
by deleting the cache file.

=back

=cut

######################################################################

sub ClearLogonParms {
    my $self = shift;

    my $file = $self->{PARMFILE};

    #Write empty file to force user prompt for new information
    &IOFileWrite($file);
    undef $self->{USER};
    &ForceParm("CQUSER", undef);
    &ForceParm("USER", undef);
    return;
}

######################################################################
#::method GetLogonParms

=over 6

=item C<GetLogonParms()>

Method that retrieves the user name and password to logon to
ClearQuest from a file stored in the user's home directory. If the
file doesn't exist, this method interacts with the user to get a
ClearQuest user name and password and stores the information into a
file for use in later sessions. This method also sets the internal
parameter "USER" for access by filters and other parameters. 

=back

=cut

######################################################################

sub GetLogonParms {
    my $self = shift;

    # This is a back door for testing only and is NOT intended
    # as a user-visible way to logon. It will NOT rely on or disturb
    # existing cached passwords, etc. 

    if ($ENV{"CQCC_TEST_USER"}) {       
        $self->{USER} = $ENV{"CQCC_TEST_USER"};
        &ForceParm("USER", $self->{USER});
        $self->{PASS} = 
            $ENV{"CQCC_TEST_PW"} ? $ENV{"CQCC_TEST_PW"} : "";
        if ($ENV{"CQCC_TEST_CQUSER"}) {
            &ForceParm("CQUSER", $ENV{"CQCC_TEST_CQUSER"});
        }
        return;
    }

    # Return if we've already been here and have User/Pass 

    if ($self->{USER}) {         
        return;
    }

    # Windows uses USERNAME, UNIX uses LOGNAME

    my $envUserName = $ENV{USERNAME};
    if (! defined $envUserName) { 
        $envUserName = $ENV{LOGNAME};
    } 

    # Read existing params file for CQ logon info

    my $parmfile = $self->{PARMFILE};
    my ($abort, $user, $pass, $username,  $cqusername);

    # Passwords are now encrypted combined with the current username
    # in order to prevent users from copying each other's cqparams
    # files. If the decrypted password/username combination doesn't
    # agree with the current username then the password is invalidated
    # and the user is asked to reenter the password information. 

    if (-s $parmfile) {
        my ($answer) = &IOFileRead($parmfile,0);
        ($user, $pass, $cqusername) = split(" ", $answer);
        if (defined $pass) {
            if ($CQCCext701 == 1) {
                my $result = 0;
                eval { 
                    require CQCCext;
                    eval {
                        $pass = CQCCext::decrypt_password($pass);
                        $result = 1;
                    };
                };
                if ($@ || $result ne 1) {
                    $pass = undef;
                    $user = undef;
                    $cqusername = undef;
                }
            } else { 
                my $decode = $pass;
                $decode =~ tr/M-Za-v0-5A-L6-9w-z/A-Za-z0-9/;
                ($pass, $username) = split("uNaMe", $decode);
                if ($username ne $envUserName) {
                    $pass = undef;
                    $cqusername = undef;
                }
            }
        }
    }

    # If not defined yet, reprompt user for information and save it

    if (! (defined $user && defined $pass)) {       
        my $notValid = 1;
        if ($ENV{"CQCC_CQ_NOLOGIN"}) {
            &IOMessage("FATAL", ["CQ_E_TEST_LOGIN_FAILED"]);
        }
        while ($notValid) { 
            # For CCRC, break up login prompt into two segments
            # and mark password prompt with special flag that will be 
            # recognized and stripped out by CCRC client Clearprompt emulator
            if (defined $ENV{ATRIA_WEB_CLEARPROMPT} && 
                ! defined $ENV{ATRIA_CTE_DYNAMIC}) {
                ($abort,$user) = 
                    &IOPrompt(["CQ_P_LOGIN_USER", "PARM:CQCC_DATABASE"]);
                if (! $abort) { 
                    ($abort,$pass) = 
                        &IOPrompt(["CQ_P_LOGIN_PASSWORD", "<{PASSWORD}>"]);
                }
            } else {
                ($abort,$user,$pass) = 
                    &IOPromptLogin(["CQ_P_LOGIN"],
                                   ["CQ_P_LOGIN_DIALOG", "PARM:CQCC_DATABASE"]);
            }
            if ($abort) { 
                &IOExitHandler(1);
            }
            elsif (defined $user && length($user) > 0 && defined $pass) {
               $notValid = 0;
            }
        }
        # remove quotes if user entered any. CQ allows quotes in password
        # so only remove begining and ending quotes for password.
        if ($pass =~ /^\"([\S\s]*)\"$/) {
            $pass = $1; 
        }
        $user =~ s/['"]//g;

        my $encrpass;
        if ($CQCCext701 == 1) {
            eval { 
                require CQCCext;
                $encrpass = CQCCext::encrypt_password($pass);
            };
            if ($@) {
                $pass = undef;
                $user = undef;
                $cqusername = undef;
            }
        } else {
            $encrpass = $pass . "uNaMe" . $envUserName;
            $encrpass =~ tr/A-Za-z0-9/M-Za-v0-5A-L6-9w-z/;
        }
        &IOFileWrite($parmfile, "$user $encrpass");
    }

    $self->{USER} = $user;
    &ForceParm("USER", $user);
    &ForceParm("CQUSER", $cqusername);
    $self->{PASS} = $pass;

    return;
}

######################################################################
#::method new

=over 6

=item C<new CQCC::CQuest()>

This method presets the object's attributes including the default
entity to be used and the name of the file used to store logon
information. During operation, additional attributes will be set as
needed including the user name and password and the query filters once
they have been parsed.

=back

=cut

######################################################################

sub new { 
   my $proto = shift;
   my $class = ref($proto) || $proto;  # Object or class name

   # Attributes
   my $self = {
       CLASSNAME => "CQCC::CQuest",

       # Configuration values
       FILTERS => undef,       # Filter list evaluated when needed by 
       # Internally set values
       PARMFILE => undef,      # File for ClearQuest logon info
       USER => undef,          # ClearQuest logon user id
       PASS => undef,          # ClearQuest logon user password
       DATABASE => undef,      # Name of database we're logged into
       @_,                     # Override previous attributes
       @_,                     # Extra time for ccperl
   };

   # In CC 6.0, rely on password file in user home directory
   # In CC 7.0.1 and beyond, rely on centrally stored secure passwords
   if ($perl5006 == 1) { 
       eval { 
           require CQCCext;
           if (defined &CQCCext::encrypt_password) {
               $CQCCext701 = 1;
           }
       };
   }
   $self->{PARMFILE} = &IOPathname(($CQCCext701 ? "PARM" : "PARM_USER"), 
                                   ".cqparams");

   bless($self, $class);
   return $self;
}

######################################################################
#::method PerformQuery 

=over 6

=item C<PerformQuery(db, entity, schema, query, results, AllowCache,
{K=>V,...})>

This method will query the ClearQuest database for records matching
the caller's query and return them as a list of strings in the format
requested by the results argument. 

Note that one time queries and results lists are supported.  The query
can be either a named query in the schema or an array reference to one
or more query clauses which will be translated using the schema
information.  The results argument can be either a named results set
or an array containing a results list and a results format.

The AllowCache argument is either TRUE or FALSE - if enabled it will 
cache query results and return saved results when the same query is run. 

The optional Key=>Value pairs allow for future expansion and include
the following: 

=over 4

RETURN => STRING (default) or ARRAY: Controls the format of the data
being returned. 

=over 4 

If STRING, each record is returned as a formatted string based on the
results format string and the results are sorted.

If ARRAY it is returned as an array reference for use with columnar
output and the first record contains the column labels. 

=back 

=back

This frontend method will check and convert input before calling the
interface specific _PerformQuery method to perform the query itself. 

=back

=cut

######################################################################

sub PerformQuery {
    my ($self, $dbname, $entity, $schema, $query, $results, 
        $cachingAllowed, @keyValuePairs) = @_;
    my $cachingEnabled = 0;
    my $returnType = "STRING";
    my $queryType = "LOCAL";  # Alternative is CQ
    my ($resultsSet, $resultsFormat);
    my @data;

    if ( $cachingAllowed eq "TRUE" && $dbname && $entity && 
        (ref($query) ne "ARRAY") && (ref($results) ne "ARRAY")) {
        $cachingEnabled = 1;
    }

    while (@keyValuePairs) { 
        my $key = shift @keyValuePairs;
        my $value = shift @keyValuePairs;
        if ($key eq "RETURN") { 
            $returnType = $value;
        }
        elsif ($key eq "QUERY_TYPE") { 
            $queryType = $value;
        }
    }

    if (! defined $schema) {
        $schema = &GetCQSchemaByEntity($entity);
        if (! defined $schema) {
            &IOMessage("ERROR", ["CQ_E_UNKNOWN_ENTITY", $entity]);
            return;
        }
    }
    
    if ( $cachingEnabled ) {
        my $key = "$dbname|$entity|$query|$results";
        my $cachedQuery = $performQueryCache{$key};
        if ($cachedQuery) {
            return @{ $cachedQuery };
        }
    }        
        
    # Internal Queries are defined in the config file and the
    # query definition and results sets need to be retrieved from
    # the local schema object. 

    my $t1 = &TimerStart("CQuest::PerformQuery($entity)");

    if ($queryType eq "LOCAL") {
        my @filters = $schema->TranslateQuery($query);
        if (! @filters) {
            &IOMessage("ERROR", ["CQ_E_UNKNOWN_QUERY", $query]);
            return;
        }

        # Get the results set description
        ($resultsSet, $resultsFormat) = 
            $schema->GetResultSet($results);
    
        # If Query doesn't have its own results, try STANDARD

        if (! defined $resultsSet || ! defined $resultsFormat) {
            ($resultsSet, $resultsFormat) = 
                $schema->GetResultSet("STANDARD");
        }

        if (! defined $resultsSet || ! defined $resultsFormat) {
            &IOMessage("ERROR", ["CQ_E_UNKNOWN_RESULTS", $results]);
            return;
        }

        &IODebugPrint("BASIC",
            "\nPerformQuery: entity $entity database $dbname\n"
            . "query $query \n    " . 
            join("\n    ",@filters),
            "\nResults: $resultsSet\n    $resultsFormat\n");
         @data = $self->_PerformQuery($dbname, $entity, $schema, 
             \@filters, $resultsSet, $resultsFormat, $returnType,
             $queryType);
    }
    # Named queries are located in the ClearQuest database and 
    # will be retrieved when needed. 

    elsif ($queryType eq "CQ") {
        &IODebugPrint("BASIC",
            "\nPerformQuery: entity $entity database $dbname\n"
		      . "named query $query\n");
        @data = $self->_PerformQuery($dbname, $entity, $schema, 
             $query, undef, undef, $returnType,$queryType);

    }

    else {
        &IOMessage("FATAL", ["CQ_E_OP_UNSUPPORTED", "CQPerlAPI",
                             "PerformQuery type $queryType"]);
        return;
    }

    &TimerStop($t1);

    if (&GetParm("CQCC_DEBUG") > 1) {
        if ($returnType eq "STRING" && $queryType eq "LOCAL") {
            my $titles = sprintf $resultsFormat, 
                             split(",",$resultsSet);
            IODebugPrint("DETAILED", "\n$titles\n" . 
                         join("\n",@data) . "\n");
        }
        else {
            for (@data) { 
                IODebugPrint("DETAILED", 
                             "[" . join("][",@_) . "]\n");
            }
        }
    }

    if ( $cachingEnabled ) {
        my $key = "$dbname|$entity|$query|$results";
        $performQueryCache{$key} = \@data;
    }        

    return @data;
}

######################################################################
#::method CheckRestrictions

=over 6

=item C<CheckRestrictions(AssocList)>

This method takes the associations in the AssocList and tests
them against the restrictions query specified for their respective
schema to make sure they are permissible as new associations. 

Returns 0 if any items fail.

=back

=cut

######################################################################

sub CheckRestrictions {
    my ($self, $assocList) = @_;
    my @entityDBlist = $assocList->GetEntityDBList();

    if ($#entityDBlist > 0) { 
        &IOMessage("ERROR", ["CQ_E_MIXED_ENTITIES",
            join(", ",@entityDBlist)]);
        return 0;
    }

    for (@entityDBlist) { 
        my ($entity, $dbname) = split(":",$_);
        my $schema = &GetCQSchemaByEntity($entity);    

        my $stdResultSet = "STANDARD"; # Standard query used
        my $multiSite = 0;             # Indicate multi-site is not enabled
        my %ccChangeSetRecs = ();      # Mapping change_set to defect recs
        my @ccChangeSets;              # List of change_sets to test 

        if (! $schema) { 
            &IOMessage("ERROR", ["CQ_E_UNKNOWN_SCHEMA", $entity]);
            return 0;
        } 
        my @query = $schema->GetRestrictionsQuery();

        # Add MASTERSHIP restrictions to the list.
        # Will make implicit Restrictions list even if one is not set
        # No harm if clause added a second time

        if (&GetParm("CQCC_MULTISITE") ne "FALSE") {
            push @query, "<RATL_MASTERSHIP> eq <*REPLICA_NAME*>";

            # Convert standard results by PRE-PENDING cc_change_set
            $multiSite = 1;
            my ($tempSet, $tempFormat) =  $schema->GetResultSet($stdResultSet);
            $tempSet = "cc_change_set," . $tempSet;
            $tempFormat = "%s " . $tempFormat;
            my @set = ($tempSet, $tempFormat);
            $stdResultSet = \@set;
        }

        if (@query) {
            my @IDs = $assocList->GetBugList();
            my @origQuery = $schema->TranslateQuery(\@query);
            push @query, "<ID> in " . join(",", @IDs);
            my @results = $self->PerformQuery($dbname, $entity,
               $schema, \@query, $stdResultSet, "FALSE");
            my (%resultsIDs, @failedIDs);

            for (@results) {
                my $record = $_;
                # Break out pre-pended cc_change_set info
                if ($multiSite eq 1) {
                    my $csetID;
                    ($csetID,$record) = split(" ",$record,2);

                    # If change_set is actively used, then 
                    # remember associated record for reporting later
                    # and add cc_change_set to list 

                    if ($csetID ne 0) {
                        $ccChangeSetRecs{$csetID} = $record; 
                        push (@ccChangeSets, $csetID);
                    }
                }
                my ($ID) = split(" ",$record);
                $resultsIDs{$ID} = 1;
            } 
            for (@IDs) { 
                if (! $resultsIDs{$_}) {
                    push @failedIDs, $_;
                }
            }

            # Return only if proposed associations pass restrictions query AND
            #    if multisite that there are no existing cc_change_sets
            
            if ($#IDs == $#results) {
                # Original records passed the test and provided all of their
                # change sets. Test mastership of these before returning. 
                # NOTE: CQWebASP can't query stateless recs like cc_change_sets

                if ($multiSite && $#ccChangeSets gt -1 &&
                    $self->{CLASSNAME} ne "CQCC::CQWeb") {
                    # Check cc_change_sets for mastership using second query 
                    if (! defined $csetSchema2) {
                        $csetSchema2 = 
                            new CQCC::CQSchema(NAME=>"cc_change_set2",
                                               TYPE=>"cc_change_set",
                                               ENTITY_LIST => "cc_change_set");
                    }
                    
                    my @csetQuery = ( "dbid in " . join(",", @ccChangeSets), 
                                      "ratl_mastership neq <*REPLICA_NAME*>");
                    my @csetResults = ("dbid,ratl_mastership", "%d,%s" );
                    my @ccChangeResults = 
                        $self->PerformQuery($dbname, "cc_change_set",
                            $csetSchema2, \@csetQuery, \@csetResults, "FALSE");
                    # Have mastership problems
                    if ($#ccChangeResults gt -1) { 
                        my @badRecords;
                        for (@ccChangeResults) {
                           my ($dbid,$replica) = split(",",$_);
                           push @badRecords, 
                               $ccChangeSetRecs{$dbid} . " [cc_change_set($replica)]";
                        }
                        # TODO: Different error message needed ?
                        IOMessage("WARN", ["CQ_E_ASSOC_NOT_ALLOWED",
                            join("\n",@badRecords), join("\n",@origQuery)]);
                        return 0;
                    } 
                }
                return 1;
            }

            # Do follow up query and show info about the wrong values
            # Relying on the STANDARD QUERY to get enough info

            my ($resultSet, $format) = 
                $schema->GetResultSet("STANDARD");

            # Append RATL_MASTERSHIP to follow up query
            if (&GetParm("CQCC_MULTISITE") ne "FALSE" &&
                !($resultSet =~ /RATL_MASTERSHIP/)) {
                    $resultSet .= ",<RATL_MASTERSHIP>";
                    $format .= " %s";
            }

            @results = $self->PerformQuery($dbname, $entity,
               $schema, [ "<ID> in " . join(",", @failedIDs) ],
               [ $resultSet, $format ],
               "FALSE");

            # If query succeeds then show information about the 
            # failed records else just show their numbers
                        
            if ($#failedIDs == $#results) { 
               IOMessage("WARN", ["CQ_E_ASSOC_NOT_ALLOWED",
                  join("\n",@results), join("\n",@origQuery)]);
            } else {
               IOMessage("WARN", ["CQ_E_ASSOC_NOT_ALLOWED",
                 join(",",@failedIDs), join("\n",@origQuery)]);
            }
        } else {
            return 1;
        }
    } 
    return 0;
}

######################################################################
#::method PrintChangeSets

=over 6

=item C<PrintChangeSets(dbid, entityName, @defectIds)>

This abstract method should be overridden by the subclass to actually
print out the cc_change_set and cc_vob_objects associated with one or
more defect records. If the information can't be obtained this default
method will report that. 

=back

=cut

######################################################################

sub PrintChangeSets { 
    IOPrint("PrintChangeSets method not supported\n");
}

######################################################################
#::method SelfTest

=over 6

=item C<SelfTest(@args)>

This method runs basic unit testing on the object instance which
currently includes connecting to the ClearQuest database, performing a
query and displaying the results. It makes new associations, queries
for current associations and deletes the new associations to confirm
operation.

=back

=cut

######################################################################

sub SelfTest {
    my $self = shift;
    my @args = @_;
    my $seed = substr(time(),2);
    my $oid = "$seed.$seed.ffff.01:02:03:04:05:06";
    my $vobID = "$seed.$seed.0000.01:02:03:04:05:06";
    my $pathName = "$self->{CLASSNAME} SelfTest";
    my $voList = "$pathName,$vobID,$oid,$oid";
    &ForceParm("CLEARCASE_PN","pathName");

    $self->Connect();

    my $s = &GetCQSchemaByEntity("defect");
    if (! defined $s) { 
        $s = &DefineCQSchema(NAME=>"Schema1",
                             ENTITY_LIST => "defect");
    } 

    my @bugQuery = 
        $self->PerformQuery(&GetParm("CQCC_DATABASE"), "defect", $s,
                            "STANDARD", "STANDARD", "TRUE");
    my $bugline;
    for $bugline (@bugQuery) {
        print $bugline,"\n";
    } 

    @bugQuery = 
        $self->PerformQuery(&GetParm("CQCC_DATABASE"), "defect", $s,
                            "STANDARD", "STANDARD", "TRUE");

    my $dbName = &GetParm("CQCC_DATABASE");
    my $entity = &GetParm("CQCC_ENTITY");

    print "\n**********\nBEFORE new associations: \n";
    my $assocList = new CQCC::AssocList();
    $assocList->ProcessBugString($entity, $dbName . "29");
    my @numbers = $assocList->GetBugList();
   
    $self->PrintChangeSets($dbName, $entity, @numbers);

    $self->EditAssocs($voList, $assocList, undef);
    
    print "\n***********\nAFTER new associations added: \n";
    $self->PrintChangeSets($dbName, $entity, @numbers);

    $self->EditAssocs($voList, undef, $assocList);

    print "\n***********\nAFTER new associations deleted: \n";
    $self->PrintChangeSets($dbName, $entity, @numbers);
    return 0;
}

######################################################################
#::method SetDatabase

=over 6

=item C<SetDatabase(databaseName)>

This method changes the database we are attached to and attempts
to login to the new database. If that fails it will try to relogin
to the original database it was logged into. If that fails it will
exit because there is no valid database. 

=back

=cut

######################################################################

sub SetDatabase {
    my ($self, $dbName) = @_;

    if ($dbName ne $self->{DATABASE}) { 
        my $oldDB = $self->{DATABASE};
        &ForceParm("CQCC_DATABASE", $dbName);
        if (! $self->Connect(1)) {
           &ForceParm("CQCC_DATABASE", $oldDB);
           $self->Connect(0);
        }
    }
}

######################################################################
#::method SetDBEntity

=over 6

=item C<SetDBEntity(database, entity)>

This method connects to the requested database and confirms that the 
entity is valid for the current CQschema. It returns 0 if there is an
error and 1 if the operation succeeds.

=back

=cut

######################################################################

sub SetDBEntity {
    my ($self, $db, $entity) = @_;

    if (! defined $db || $db eq "" || ! defined $entity || $entity eq "") {
        return 0;
    }
    if ($db ne &GetParm("CQCC_DATABASE")) {
       # Would need to relogin to database
       $self->SetDatabase($db);
       if ($db ne &GetParm("CQCC_DATABASE")) {
           &IOMessage("FATAL", ["CQ_E_UNKNOWN_DATABASE", $db]);
           return 0;
       }
    }
    if ($entity ne &GetParm("CQCC_ENTITY")) {
        my @entList = &SupportedEntityList(&GetParm("CQCC_DATABASE"));
        my $found;
        for (@entList) { 
            if ($_ eq $entity) {
               $found = 1;
            }
        }
        if (! $found) {
           &IOMessage("FATAL", ["CQ_E_UNKNOWN_ENTITY", $entity]);
           return 0;
        } else {
           &ForceParm("CQCC_ENTITY", $entity);
        }
    }
    return 1;
}

######################################################################
#::method ValidateDatabaseEntity

=over 6

=item C<ValidateDatabaseEntity()>

This method validates the current settings for CQCC_DATABASE 
and CQCC_ENTITY to ensure that the values are self consistent
and are on the list of valid database/entity combinations. It
is usually called after the main trigger loads any prior settings
from the cache file and has read in command line arguments. It
will not handle changing database logins. 

=back

=cut

######################################################################

sub ValidateDatabaseEntity {
    my ($db, $entity, $onList, $schema);

    # Behavior of database/entity settings
    # (1) If trigger -entity argument sets default, that presets var
    # (2) If cache-var defined, vars were forced on Base::LoadCache
    # (3) Else values preferences can be set from ENV or Config 
    # (4) Else they are unset and defaults will be determined. 

    # Confirm CQCC_DATABASE is a known database else 
    # set it to the first element in the database list

    my @dbList = SupportedDatabaseList();
    $onList = 0;
    if ($db = GetParm("CQCC_DATABASE")) {
       for (@dbList) {
          if ($db eq $_) {
             $onList = 1;
             last;
          }
       }
    }
    if (! (defined $db && $onList)) {
       ($db) = @dbList;
       &ForceParm("CQCC_DATABASE", $db);
    }
   
    # Confirm CQCC_ENTITY is valid for the current database
    
    my @entList = SupportedEntityList($db);
    $onList = 0;
    if ($entity = GetParm("CQCC_ENTITY")) {
       for (@entList) {
          if ($entity eq $_) {
             $onList = 1;
             last;
          }
       }
    }

    if (! (defined $entity && $onList)) {
       ($entity) = @entList;
       &ForceParm("CQCC_ENTITY", $entity);
   }
 
   # Test to make sure we have a schema for the entity
   #  else warning.

   $schema = &GetCQSchemaByEntity($entity);
   if (! defined $schema) {
       &IOMessage("FATAL", ["CQ_E_NO_VALID_SCHEMA", $db, $entity]);
   }
}

######################################################################
#::Tail
######################################################################

1;
